from typing import Any, Dict, List
from colorama import Fore, Style
from dataclasses import dataclass, asdict
import copy
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand
from promptrix.SystemMessage import SystemMessage
from llmsearch import search_service

search_schema = CommandSchema(
    schema_type="object",
    title="search",
    description="web search for provided query",
    properties={
        "query": {
            "type": "string",
            "description": "web query string"
            }
        }
    required: {'query']
    }
)


class SearchCommand(SchemaBasedCommand):
    def __init__(self, client, model, name, description=None):
        super().__init__(search_schema, search_schema.title, search_schema.description)


    def execute(self, input: input, memory: Any, functions: Any, tokenizer: Any) -> Any:
        try:
            print(f'***** SearchCommand input {input}')
            if type(input) is dict and 'query' in dict:
                query = input['query']
                response = search_service.run_chat(query)
                print(f'***** SearchCommand response {response}')
                return {'status':'success', 'message':response}
        except Exception as err:
            message = str(err)
            print(f'***** SearchCommand error {str{e}')
            return asdict(TaskResponse('TaskResponse', 'error', message))



import argparse
import json
import traceback
import socket
import requests
from alphawave_pyexts.conversation import Conversation, SeparatorStyle
import alphawave_pyexts.conversation as cv
import tkinter as tk
import time

MODEL_NAME = None; WORKER_ADDR=None; CONTROLLER_ADDRESS = "http://localhost:21001"
USER_PREFIX = ''
ASSISTANT_PREFIX = ''

host='192.168.1.195'
port = 5004
cv.register_conv_template(Conversation(
        name="falcon_instruct",
        system="",
        roles=("HUMAN", "ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
)
cv.register_conv_template(Conversation(
        name="falcon_instruct2",
        system="",
        roles=(" ###HUMAN", " ###ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
)
cv.register_conv_template(Conversation(
        name="falcon_instruct3",
        system="",
        roles=("HUMAN", "ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
)
cv.register_conv_template(Conversation(
        name="wizardLM",
        system="",
        roles=("### HUMAN", "### ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
)

def run_query(model, messages, temp, top_p, max_tokens, host = host, port = port, tkroot = None, tkdisplay=None): 
    global USER_PREFIX, ASSISTANT_PREFIX
    conv=cv.get_conv_template(model)
    USER_PREFIX = conv.roles[0]
    ASSISTANT_PREFIX = conv.roles[1]
    # set this so client can check for run-on, although this test should pbly be here!
    for msg in messages:
        #print(f'  {msg}')
        role = msg['role']
        if role.lower() == 'user':
            role_index = 0
        else:
            role_index = 1
        conv.append_message(conv.roles[role_index], msg['content'])
    conv.append_message(conv.roles[1], '')
    prompt = conv.get_prompt()
    server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens}
    """
    prompt = ''
    for msg in messages:
        if not isinstance(msg, dict):
            msg = msg.__dict__
        role = msg['role']
        if role.lower() == 'user' or role.lower() == 'system':
            role_os = USER
        elif role.lower() == 'assistant':
            role_os = ASSISTANT
        else: print(f'***** unknown role {role}')
        prompt += role_os + str(msg['content'])
    prompt += ASSISTANT
    """
    server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens}
    smj = json.dumps(server_message)
    try:
        client_socket = socket.socket()  # instantiate
        client_socket.connect((host, port))  # connect to the server
        client_socket.settimeout(240)
        server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens}
        smj = json.dumps(server_message)
        client_socket.sendall(smj.encode('utf-8'))
        client_socket.sendall(b'x00xff')
        response = ''
        while True:
            s = client_socket.recv(1024) # break every 32 chars to stream output
            if s is None or not s:
                break
            if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                s = s[:-6].decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
                break
            else:
                s = s.decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
        client_socket.close()  # close the connection
        #check for run on hallucination in response 
        runon_idx = response.find(conv.roles[0])
        if runon_idx > 0:
            response = response[:runon_idx]
        return response
    except: 
        traceback.print_exc()
    return ''
    



from pyee import AsyncIOEventEmitter
from jsonschema import validate, ValidationError
from typing import Any, Dict, Optional, Union
import json
from promptrix.promptrixTypes import PromptMemory, PromptFunctions, Tokenizer
from dataclasses import dataclass, asdict
import traceback

@dataclass
class CommandSchema:
    def __init__(self, type: str, title: str='', description: str='', returns: Optional[str] = None):
        self.type = type
        self.title = title
        self.description = description
        self.returns = returns

class SchemaBasedCommand(AsyncIOEventEmitter):
    def __init__(self, schema: CommandSchema, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__()
        self._schema = schema
        self._title = title
        self._description = description

    @property
    def description(self) -> str:
        return self._description or self._schema.description

    @property
    def inputs(self) -> Optional[str]:
        if self._schema.properties:
            properties = self._schema.properties or {}
            inputs = [f'"{key}":"<{property.get("description", property.get("type", "any"))}>"' for key, property in properties.items()]
            return ",".join(inputs)
        else:
            return None

    @property
    def output(self) -> Optional[str]:
        return self._schema.returns

    @property
    def schema(self) -> CommandSchema:
        return self._schema

    @property
    def title(self) -> str:
        return self._title or self._schema.title

    async def execute(self, input: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer') -> Any:
        raise NotImplementedError

    async def validate(self, input: Dict[str, Any], memory: 'PromptMemory', functions: 'PromptFunctions', tokenizer: 'Tokenizer') -> 'Validation':
        try:
            cleaned = self.clean_input(input)
            if self._schema is not None:
                validate(cleaned, asdict(self._schema))
            return {
                'type': 'Validation',
                'valid': True,
                'value': cleaned
            }
        except ValidationError as e:
            errors = f'"{e.path[0] if e.path else "input"}": {e.message}'
            message = errors
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': f"The ['command']['input'] field has errors:\n{message}\n\nTry again."
            }

    def clean_input(self, input: Dict[str, Any]) -> Dict[str, Any]:
        cleaned = {}
        properties = self._schema.properties
        if not properties:
            return cleaned

        for key, property in properties.items():
            type = property.get('type', 'any')
            value = input.get(key)

            if value is None or (isinstance(value, str) and value.startswith("<") and value.endswith(">")):
                continue

            if not isinstance(value, str):
                value = json.dumps(value) if isinstance(value, dict) else str(value)

            try:
                if type == 'string':
                    cleaned[key] = str(value)
                elif type == 'number':
                    cleaned[key] = float(value)
                elif type == 'boolean':
                    cleaned[key] = bool(value)
                elif type in ['array', 'object']:
                    cleaned[key] = json.loads(value)
            except Exception:
                pass

        return cleaned
